@@ -169,6 +169,12 @@ class DocumentSchema(object):
             self._doc['_id'] = value
         elif key == "_deleted":
             self._doc["_deleted"] = value
+        elif key == "_attachments":
+            if key not in self._doc or not value:
+                self._doc[key] = {}
+            elif not isinstance(self._doc[key], dict):
+                self._doc[key] = {}
+            value = LazyDict(self._doc[key], init_vals=value)
         else:
             check_reserved_words(key)
             if not hasattr( self, key ) and not self._allow_dynamic_properties:
@@ -227,7 +233,10 @@ class DocumentSchema(object):
             return self._dynamic_properties[key]
         elif key  in ('_id', '_rev', '_attachments', 'doc_type'):
             return self._doc.get(key)
-        return self.__dict__[key]
+        try:
+            return self.__dict__[key]
+        except KeyError, e:
+            raise AttributeError(e)
 
     def __getitem__(self, key):
         """ get property value
@@ -325,6 +334,7 @@ class DocumentSchema(object):
                     value = value_to_python(value)
                     setattr(instance, attr_name, value)
         return instance
+    from_json = wrap
 
     def validate(self, required=True):
         """ validate a document """
@@ -336,16 +346,9 @@ class DocumentSchema(object):
 
     def clone(self, **kwargs):
         """ clone a document """
-        for prop_name in self._properties.keys():
-            try:
-                kwargs[prop_name] = self._doc[prop_name]
-            except KeyError:
-                pass
-
         kwargs.update(self._dynamic_properties)
         obj = self.__class__(**kwargs)
         obj._doc = self._doc
-
         return obj
 
     @classmethod
@@ -446,13 +449,12 @@ class DocumentBase(DocumentSchema):
         be committed even if this creates conflicts.
 
         """
-        if cls._db is None:
-            raise TypeError("doc database required to save document")
-        docs_to_save= [doc._doc for doc in docs if doc._doc_type == cls._doc_type]
+        db = cls.get_db()
+        docs_to_save= [doc for doc in docs if doc._doc_type == cls._doc_type]
         if not len(docs_to_save) == len(docs):
             raise ValueError("one of your documents does not have the correct type")
-        cls._db.bulk_save(docs_to_save, use_uuids=use_uuids, all_or_nothing=all_or_nothing)
-    
+        db.bulk_save(docs_to_save, use_uuids=use_uuids, all_or_nothing=all_or_nothing)
+
     bulk_save = save_docs
 
     @classmethod
@@ -469,11 +471,9 @@ class DocumentBase(DocumentSchema):
         """ get  or create document with `docid` """
        
         if db:
-            cls._db = db
+            cls.set_db(db)
         cls._allow_dynamic_properties = dynamic_properties
-
-        if cls._db is None:
-            raise TypeError("doc database required to save document")
+        db = cls.get_db()
 
         if docid is None:
             obj = cls()
@@ -483,7 +483,7 @@ class DocumentBase(DocumentSchema):
         rev = params.pop('rev', None)
 
         try:
-            return cls._db.get(docid, rev=rev, wrapper=cls.wrap, **params)
+            return db.get(docid, rev=rev, wrapper=cls.wrap, **params)
         except ResourceNotFound:
             obj = cls()
             obj._id = docid
