@@ -10,16 +10,28 @@ import datetime
 import re
 import time
 
+try:
+    from collections import MutableSet, Iterable
+
+    def is_iterable(c):
+        return isinstance(c, Iterable)
+except ImportError:
+    from sets import Set as MutableSet
+    
+    def is_iterable(o):
+        return hasattr(c, '__iter__')
+
 from couchdbkit.exceptions import BadValueError 
 
 __all__ = ['ALLOWED_PROPERTY_TYPES', 'Property', 'StringProperty',
         'IntegerProperty', 'DecimalProperty', 'BooleanProperty',
         'FloatProperty', 'DateTimeProperty', 'DateProperty',
         'TimeProperty', 'DictProperty', 'ListProperty',
-        'StringListProperty', 'dict_to_json', 'list_to_json',
+        'StringListProperty', 'SetProperty',
+        'dict_to_json', 'list_to_json',
         'value_to_json', 'MAP_TYPES_PROPERTIES', 'value_to_python',
         'dict_to_python', 'list_to_python', 'convert_property',
-        'value_to_property', 'LazyDict', 'LazyList']
+        'value_to_property', 'LazyDict', 'LazyList', 'LazySet']
 
 ALLOWED_PROPERTY_TYPES = set([
     basestring,
@@ -35,6 +47,7 @@ ALLOWED_PROPERTY_TYPES = set([
     decimal.Decimal,
     dict,
     list,
+    set,
     type(None)
 ])
 
@@ -512,6 +525,86 @@ class StringListProperty(ListProperty):
         super(StringListProperty, self).__init__(verbose_name=verbose_name,
             default=default, required=required, item_type=basestring, **kwds)
 
+
+class SetProperty(Property):
+    """A property that stores a Python set as a list of unique
+    elements.
+
+    Note that Python set operations like union that return a set
+    object do not alter list that will be stored with the next save,
+    while operations like update that change a set object in-place do
+    keep the list in sync.
+    """
+    def __init__(self, verbose_name=None, default=None, required=None,
+                 item_type=None, **kwds):
+        """Construct SetProperty.
+
+         :args verbose_name: Optional verbose name.
+
+         :args default: Optional default value; if omitted, an empty
+                        set is used.
+
+         :args required: True if field is required, default is False.
+
+         :args item_type: Optional data type of items that set
+                          contains.  Used to assist with JSON
+                          serialization/deserialization when data is
+                          stored/retireved.
+
+         :args **kwds: Optional additional keyword arguments, passed to
+                       base class.
+         """
+        if default is None:
+            default = set()
+        if item_type is not None and item_type not in ALLOWED_PROPERTY_TYPES:
+            raise ValueError('item_type %s not in %s'
+                             % (item_type, ALLOWED_PROPERTY_TYPES))
+        self.item_type = item_type
+        super(SetProperty, self).__init__(
+            verbose_name=verbose_name, default=default, required=required,
+            **kwds)
+
+    data_type = set
+
+    def validate(self, value, required=True):
+        value = super(SetProperty, self).validate(value, required=required)
+        if value and value is not None:
+            if not isinstance(value, MutableSet):
+                raise BadValueError('Property %s must be a set' % self.name)
+            value = self.validate_set_contents(value)
+        return value
+
+    def validate_set_contents(self, value):
+        try:
+            value = validate_set_content(value, item_type=self.item_type)
+        except BadValueError:
+            raise BadValueError(
+                'Items of %s set must all be in %s' %
+                    (self.name, ALLOWED_PROPERTY_TYPES))
+        return value
+
+    def default_value(self):
+        """Return default value for set.
+
+        Because the property supplied to 'default' is a static value,
+        that value must be shallow copied to prevent all fields with
+        default values from sharing the same instance.
+
+        Returns:
+          Copy of the default value.
+        """
+        value = super(SetProperty, self).default_value()
+        if value is None:
+            return set()
+        return value.copy()
+
+    def to_python(self, value):
+        return LazySet(value, item_type=self.item_type)
+
+    def to_json(self, value):
+        return value_to_json(value, item_type=self.item_type)
+
+
 # structures proxy
 
 class LazyDict(dict):
@@ -670,10 +763,23 @@ class LazyList(list):
             self.doc.append(value_to_json(value, item_type=self.item_type))
         super(LazyList, self).append(value)
 
+    def extend(self, x):
+        self.doc.extend(
+            [value_to_json(v, item_type=self.item_type) for v in x])
+        super(LazyList, self).extend(x)
+
     def index(self, x, *args):
         x = value_to_json(x, item_type=self.item_type)
         return self.doc.index(x)
 
+    def insert(self, i, x):
+        self.__setslice__(i, i, [x])
+
+    def pop(self, i=-1):
+        del self.doc[i]
+        v = super(LazyList, self).pop(i)
+        return value_to_python(v, item_type=self.item_type)
+
     def remove(self, x):
         del self[self.index(x)]
 
@@ -686,6 +792,130 @@ class LazyList(list):
         list.reverse(self)
 
 
+class LazySet(MutableSet):
+    """Object to make sure that we keep set and _doc synchronized.
+
+    We sub-class MutableSet and maintain changes in doc.
+
+    Note that methods like union that return a set object do not
+    alter _doc, while methods like update that change a set object
+    in-place do keep _doc in sync.
+    """
+    def _map_named_operation(opname):
+        fn = getattr(MutableSet, opname)
+        if hasattr(fn, 'im_func'):
+            fn = fn.im_func
+        def method(self, other, fn=fn):
+            if not isinstance(other, MutableSet):
+                other = self._from_iterable(other)
+            return fn(self, other)
+        return method
+
+    issubset = _map_named_operation('__le__')
+    issuperset = _map_named_operation('__ge__')
+    symmetric_difference = _map_named_operation('__xor__')
+
+    def __init__(self, doc, item_type=None):
+        self.item_type = item_type
+        self.doc = doc
+        self.elements = set(value_to_python(value, self.item_type)
+                            for value in self.doc)
+
+    def __repr__(self):
+        return '%s(%r)' % (type(self).__name__, list(self))
+
+    @classmethod
+    def _from_iterable(cls, it):
+        return cls(it)
+
+    def __iand__(self, iterator):
+        for value in (self.elements - iterator):
+            self.elements.discard(value)
+        return self
+
+    def __iter__(self):
+        return iter(element for element in self.elements)
+
+    def __len__(self):
+        return len(self.elements)
+
+    def __contains__(self, item):
+        return item in self.elements
+
+    def __xor__(self, other):
+        if not isinstance(other, MutableSet):
+            if not is_iterable(Other):
+                return NotImplemented
+            other = self._from_iterable(other)
+        return (self.elements - other) | (other - self.elements)
+
+    def __gt__(self, other):
+        if not isinstance(other, MutableSet):
+            return NotImplemented
+        return other < self.elements
+
+    def __ge__(self, other):
+        if not isinstance(other, MutableSet):
+            return NotImplemented
+        return other <= self.elements
+
+    def __ne__(self, other):
+        return not (self.elements == other)
+
+    def add(self, value):
+        self.elements.add(value)
+        if value not in self.doc:
+            self.doc.append(value_to_json(value, item_type=self.item_type))
+
+    def copy(self):
+        return self.elements.copy()
+
+    def difference(self, other, *args):
+        return self.elements.difference(other, *args)
+
+    def difference_update(self, other, *args):
+        for value in other:
+            self.discard(value)
+        for arg in args:
+            self.difference_update(arg)
+
+    def discard(self, value):
+        self.elements.discard(value)
+        try:
+            self.doc.remove(value)
+        except ValueError:
+            pass
+
+    def intersection(self, other, *args):
+        return self.elements.intersection(other, *args)
+
+    def intersection_update(self, other, *args):
+        if not isinstance(other, MutableSet):
+            other = set(other)
+        for value in self.elements - other:
+            self.discard(value)
+        for arg in args:
+            self.intersection_update(arg)
+
+    def symmetric_difference_update(self, other):
+        if not isinstance(other, MutableSet):
+            other = set(other)
+        for value in other:
+            if value in self.elements:
+                self.discard(value)
+            else:
+                self.add(value)
+
+    def union(self, other, *args):
+        return self.elements.union(other, *args)
+
+    def update(self, other, *args):
+        self.elements.update(other, *args)
+        for element in self.elements:
+            if element not in self.doc:
+                self.doc.append(
+                    value_to_json(element, item_type=self.item_type))
+
 # some mapping
 
 MAP_TYPES_PROPERTIES = {
@@ -700,7 +930,8 @@ MAP_TYPES_PROPERTIES = {
         long: LongProperty,
         float: FloatProperty,
         list: ListProperty,
-        dict: DictProperty
+        dict: DictProperty,
+        set: SetProperty,
 }
 
 def convert_property(value):
@@ -730,6 +961,10 @@ def validate_dict_content(value, item_type=None):
     return dict([(k, validate_content(v,
                 item_type=item_type)) for k, v in value.iteritems()])
 
+def validate_set_content(value, item_type=None):
+    """ validate type of values in a set """
+    return set(validate_content(item, item_type=item_type) for item in value)
+
 def validate_content(value, item_type=None):
     """ validate a value. test if value is in supported types """
     if isinstance(value, list):
@@ -766,7 +1001,7 @@ def value_to_json(value, item_type=None):
         value = value.replace(microsecond=0).isoformat()
     elif isinstance(value, decimal.Decimal) and is_type_ok(item_type, decimal.Decimal):
         value = unicode(value)
-    elif isinstance(value, list):
+    elif isinstance(value, (list, MutableSet)):
         value = list_to_json(value, item_type)
     elif isinstance(value, dict):
         value = dict_to_json(value, item_type)
@@ -782,11 +1017,11 @@ def value_to_python(value, item_type=None):
     """
     data_type = None
     if isinstance(value, basestring):
-        if re_date.match(value) and is_type_ok(item_type, datetime.datetime):
+        if re_date.match(value) and is_type_ok(item_type, datetime.date):
             data_type = datetime.date
-        elif re_time.match(value) and is_type_ok(item_type, datetime.date):
+        elif re_time.match(value) and is_type_ok(item_type, datetime.time):
             data_type = datetime.time
-        elif re_datetime.match(value) and is_type_ok(item_type, datetime.time):
+        elif re_datetime.match(value) and is_type_ok(item_type, datetime.datetime):
             data_type = datetime.datetime
         elif re_decimal.match(value) and is_type_ok(item_type, decimal.Decimal):
             data_type = decimal.Decimal
@@ -797,7 +1032,7 @@ def value_to_python(value, item_type=None):
                 value = prop.to_python(value)
             except:
                 pass
-    elif isinstance(value, list):
+    elif isinstance(value, (list, MutableSet)):
         value = list_to_python(value, item_type=item_type)
     elif isinstance(value, dict):
         value = dict_to_python(value, item_type=item_type)
