@@ -6,7 +6,10 @@
 __author__ = 'benoitc@e-engura.com (Benoît Chesneau)'
 
 import copy
-import unittest
+try:
+    import unittest2 as unittest
+except ImportError:
+    import unittest
 
 from couchdbkit import ResourceNotFound, RequestFailed, \
 ResourceConflict
