@@ -7,7 +7,10 @@ __author__ = 'benoitc@e-engura.com (Benoît Chesneau)'
 
 import datetime
 import decimal
-import unittest
+try:
+    import unittest2 as unittest
+except ImportError:
+    import unittest
 
 from couchdbkit import *
 
@@ -310,6 +313,23 @@ class DocumentTestCase(unittest.TestCase):
 
         self.assert_(len(doc2._dynamic_properties) == 1)
 
+    def testClone(self):
+        class A(DocumentSchema):
+            s = StringProperty()
+
+        class B(Document):
+            a = SchemaProperty(A)
+            s1 = StringProperty()
+
+        b = B()
+        b.s1 = "test1"
+        b.a.s = "test"
+        b1 = b.clone()
+
+        self.assert_(b1.s1 == "test1")
+        self.assert_('s' in b1._doc['a'])
+        self.assert_(b1.a.s == "test")
+
     def testView(self):
         class TestDoc(Document):
             field1 = StringProperty()
@@ -381,7 +401,6 @@ class DocumentTestCase(unittest.TestCase):
         self.assert_(results[0].__class__ == A)
         self.assert_(results[1].__class__ == B)
         self.server.delete_db('couchdbkit_test')
-
     
     def testViewNoneValue(self):
         class TestDoc(Document):
@@ -976,7 +995,285 @@ class PropertyTestCase(unittest.TestCase):
         b1 = B.get(b._id)
         self.assert_(len(b1.slm) == 2)
         self.assert_(b1.slm[0].s == "test")
-        
+
+
+    def testSchemaListPropertySlice(self):
+        """SchemaListProperty slice methods
+        """
+        class A(DocumentSchema):
+            s = StringProperty()
+            
+        class B(Document):
+            slm = SchemaListProperty(A)
+
+        b = B()
+        a1 = A()
+        a1.s = 'test1'
+        a2 = A()
+        a2.s = 'test2'
+        a3 = A()
+        a3.s = 'test3'
+        b.slm[0:1] = [a1, a2]
+        self.assertEqual(len(b.slm), 2)
+        self.assertEqual([b.slm[0].s, b.slm[1].s], [a1.s, a2.s])
+        self.assertEqual(b._doc, {
+            'doc_type': 'B',
+            'slm': [{'doc_type': 'A', 's': unicode(a1.s)},
+                    {'doc_type': 'A', 's': unicode(a2.s)}]
+        })
+        b.slm.append(a3)
+        c = b.slm[1:3]
+        self.assertEqual(len(c), 2)
+        self.assertEqual([c[0].s, c[1].s], [a2.s, a3.s])
+        del b.slm[1:3]
+        self.assertEqual(len(b.slm), 1)
+        self.assertEqual(b.slm[0].s, a1.s)
+        self.assertEqual(b._doc, {
+            'doc_type': 'B',
+            'slm': [{'doc_type': 'A', 's': unicode(a1.s)}]
+        })
+
+
+    def testSchemaListPropertyContains(self):
+        """SchemaListProperty contains method
+        """
+        class A(DocumentSchema):
+            s = StringProperty()
+            
+        class B(Document):
+            slm = SchemaListProperty(A)
+
+        b = B()
+        a1 = A()
+        a1.s = 'test1'
+        a2 = A()
+        a2.s = 'test2'
+        b.slm = [a1]
+        self.assertTrue(a1 in b.slm)
+        self.assertFalse(a2 in b.slm)
+
+
+    def testSchemaListPropertyCount(self):
+        """SchemaListProperty count method
+        """
+        class A(DocumentSchema):
+            s = StringProperty()
+            
+        class B(Document):
+            slm = SchemaListProperty(A)
+
+        b = B()
+        a1 = A()
+        a1.s = 'test1'
+        a2 = A()
+        a2.s = 'test2'
+        b.slm = [a1, a2, a1]
+        self.assertEqual(b.slm.count(a1), 2)
+
+
+    def testSchemaListPropertyExtend(self):
+        """SchemaListProperty extend method
+        """
+        class A(DocumentSchema):
+            s = StringProperty()
+            
+        class B(Document):
+            slm = SchemaListProperty(A)
+
+        b = B()
+        a1 = A()
+        a1.s = 'test1'
+        a2 = A()
+        a2.s = 'test2'
+        b.slm.extend([a1, a2])
+        self.assertEqual(len(b.slm), 2)
+        self.assertEqual([b.slm[0].s, b.slm[1].s], [a1.s, a2.s])
+        self.assertEqual(b._doc, {
+            'doc_type': 'B',
+            'slm': [{'doc_type': 'A', 's': unicode(a1.s)},
+                    {'doc_type': 'A', 's': unicode(a2.s)}]
+        })
+
+
+    def testSchemaListPropertyIndex(self):
+        """SchemaListProperty index method
+        """
+        class A(DocumentSchema):
+            s = StringProperty()
+            
+        class B(Document):
+            slm = SchemaListProperty(A)
+
+        b = B()
+        a1 = A()
+        a1.s = 'test1'
+        a2 = A()
+        a2.s = 'test2'
+        a3 = A()
+        a3.s = 'test3'
+        b.slm = [a1, a2, a1, a2, a1]
+        self.assertEqual(b.slm.index(a1), 0)
+        self.assertEqual(b.slm.index(a2, 2), 3)
+        self.assertEqual(b.slm.index(a1, 1, 3), 2)
+        self.assertEqual(b.slm.index(a1, 1, -2), 2)
+        with self.assertRaises(ValueError) as cm:
+            b.slm.index(a3)
+        self.assertEqual(str(cm.exception), 'list.index(x): x not in list')
+
+
+    def testSchemaListPropertyInsert(self):
+        """SchemaListProperty insert method
+        """
+        class A(DocumentSchema):
+            s = StringProperty()
+            
+        class B(Document):
+            slm = SchemaListProperty(A)
+
+        b = B()
+        a1 = A()
+        a1.s = 'test1'
+        a2 = A()
+        a2.s = 'test2'
+        a3 = A()
+        a3.s = 'test3'
+        b.slm = [a1, a3]
+        b.slm.insert(1, a2)
+        self.assertEqual(len(b.slm), 3)
+        self.assertEqual(
+            [b.slm[0].s, b.slm[1].s, b.slm[2].s], [a1.s, a2.s, a3.s])
+        self.assertEqual(b._doc, {
+            'doc_type': 'B',
+            'slm': [{'doc_type': 'A', 's': unicode(a1.s)},
+                    {'doc_type': 'A', 's': unicode(a2.s)},
+                    {'doc_type': 'A', 's': unicode(a3.s)}]
+        })
+
+
+    def testSchemaListPropertyPop(self):
+        """SchemaListProperty pop method
+        """
+        class A(DocumentSchema):
+            s = StringProperty()
+            
+        class B(Document):
+            slm = SchemaListProperty(A)
+
+        b = B()
+        a1 = A()
+        a1.s = 'test1'
+        a2 = A()
+        a2.s = 'test2'
+        a3 = A()
+        a3.s = 'test3'
+        b.slm = [a1, a2, a3]
+        v = b.slm.pop()
+        self.assertEqual(v.s, a3.s)
+        self.assertEqual(len(b.slm), 2)
+        self.assertEqual([b.slm[0].s, b.slm[1].s], [a1.s, a2.s])
+        self.assertEqual(b._doc, {
+            'doc_type': 'B',
+            'slm': [{'doc_type': 'A', 's': unicode(a1.s)},
+                    {'doc_type': 'A', 's': unicode(a2.s)}]
+        })
+        v = b.slm.pop(0)
+        self.assertEqual(v.s, a1.s)
+        self.assertEqual(len(b.slm), 1)
+        self.assertEqual(b.slm[0].s, a2.s)
+        self.assertEqual(b._doc, {
+            'doc_type': 'B',
+            'slm': [{'doc_type': 'A', 's': unicode(a2.s)}]
+        })
+
+
+    def testSchemaListPropertyRemove(self):
+        """SchemaListProperty remove method
+        """
+        class A(DocumentSchema):
+            s = StringProperty()
+            
+        class B(Document):
+            slm = SchemaListProperty(A)
+
+        b = B()
+        a1 = A()
+        a1.s = 'test1'
+        a2 = A()
+        a2.s = 'test2'
+        b.slm = [a1, a2]
+        b.slm.remove(a1)
+        self.assertEqual(len(b.slm), 1)
+        self.assertEqual(b.slm[0].s, a2.s)
+        self.assertEqual(b._doc, {
+            'doc_type': 'B',
+            'slm': [{'doc_type': 'A', 's': unicode(a2.s)}]
+        })
+        with self.assertRaises(ValueError) as cm:
+            b.slm.remove(a1)
+        self.assertEqual(str(cm.exception), 'list.remove(x): x not in list')
+
+
+    def testSchemaListPropertyReverse(self):
+        """SchemaListProperty reverse method
+        """
+        class A(DocumentSchema):
+            s = StringProperty()
+            
+        class B(Document):
+            slm = SchemaListProperty(A)
+
+        b = B()
+        a1 = A()
+        a1.s = 'test1'
+        a2 = A()
+        a2.s = 'test2'
+        b.slm = [a1, a2]
+        b.slm.reverse()
+        self.assertEqual([b.slm[0].s, b.slm[1].s], [a2.s, a1.s])
+        self.assertEqual(b._doc, {
+            'doc_type': 'B',
+            'slm': [{'doc_type': 'A', 's': unicode(a2.s)},
+                    {'doc_type': 'A', 's': unicode(a1.s)}]
+        })
+
+
+    def testSchemaListPropertySort(self):
+        """SchemaListProperty sort method
+        """
+        class A(DocumentSchema):
+            s = StringProperty()
+            
+        class B(Document):
+            slm = SchemaListProperty(A)
+
+        b = B()
+        a1 = A()
+        a1.s = 'test1'
+        a2 = A()
+        a2.s = 'test2'
+        b.slm = [a2, a1]
+        b.slm.sort(key=lambda item: item['s'])
+        self.assertEqual([b.slm[0].s, b.slm[1].s], [a1.s, a2.s])
+        self.assertEqual(b._doc, {
+            'doc_type': 'B',
+            'slm': [{'doc_type': 'A', 's': unicode(a1.s)},
+                    {'doc_type': 'A', 's': unicode(a2.s)}]
+        })
+        b.slm.sort(key=lambda item: item['s'], reverse=True)
+        self.assertEqual([b.slm[0].s, b.slm[1].s], [a2.s, a1.s])
+        self.assertEqual(b._doc, {
+            'doc_type': 'B',
+            'slm': [{'doc_type': 'A', 's': unicode(a2.s)},
+                    {'doc_type': 'A', 's': unicode(a1.s)}]
+        })
+        b.slm.sort(cmp=lambda x, y: cmp(x['s'].lower(), y['s'].lower()))
+        self.assertEqual([b.slm[0].s, b.slm[1].s], [a1.s, a2.s])
+        self.assertEqual(b._doc, {
+            'doc_type': 'B',
+            'slm': [{'doc_type': 'A', 's': unicode(a1.s)},
+                    {'doc_type': 'A', 's': unicode(a2.s)}]
+        })
+
 
     def testSchemaDictProperty(self):
         class A(DocumentSchema):
@@ -1047,7 +1344,6 @@ class PropertyTestCase(unittest.TestCase):
         self.assert_(b.l == [])
         self.assert_(b.to_json()['l'] == [])
         
-       
         
     def testListPropertyNotEmpty(self):
         from datetime import datetime
@@ -1076,7 +1372,8 @@ class PropertyTestCase(unittest.TestCase):
         a2 = A2()            
         self.assertTrue(a2.validate(required=False))
         self.assertTrue(a2.validate())
-    
+
+
     def testListPropertyWithType(self):
         from datetime import datetime
         class A(Document):
@@ -1102,6 +1399,106 @@ class PropertyTestCase(unittest.TestCase):
         b1.ls.remove(u'hello')
         self.assert_(u'hello' not in b1.ls)
 
+
+    def testListPropertyExtend(self):
+        """list extend method for property w/o type
+        """
+        class A(Document):
+            l = ListProperty()
+
+        a = A()
+        a.l.extend([42, 24])
+        self.assert_(a.l == [42, 24])
+        self.assert_(a._doc == {'doc_type': 'A', 'l': [42, 24]})
+
+
+    def testListPropertyExtendWithType(self):
+        """list extend method for property w/ type
+        """
+        from datetime import datetime
+        class A(Document):
+            l = ListProperty(item_type=datetime)
+
+        a = A()
+        d1 = datetime(2011, 3, 11, 21, 31, 1)
+        d2 = datetime(2011, 11, 3, 13, 12, 2)
+        a.l.extend([d1, d2])
+        self.assert_(a.l == [d1, d2])
+        self.assert_(a._doc == {
+            'doc_type': 'A',
+            'l': ['2011-03-11T21:31:01Z', '2011-11-03T13:12:02Z']
+        })
+
+
+    def testListPropertyInsert(self):
+        """list insert method for property w/o type
+        """
+        class A(Document):
+            l = ListProperty()
+
+        a = A()
+        a.l = [42, 24]
+        a.l.insert(1, 4224)
+        self.assertEqual(a.l, [42, 4224, 24])
+        self.assertEqual(a._doc, {'doc_type': 'A', 'l': [42, 4224, 24]})
+
+
+    def testListPropertyInsertWithType(self):
+        """list insert method for property w/ type
+        """
+        from datetime import datetime
+        class A(Document):
+            l = ListProperty(item_type=datetime)
+
+        a = A()
+        d1 = datetime(2011, 3, 11, 21, 31, 1)
+        d2 = datetime(2011, 11, 3, 13, 12, 2)
+        d3 = datetime(2010, 1, 12, 3, 2, 3)
+        a.l = [d1, d3]
+        a.l.insert(1, d2)
+        self.assertEqual(a.l, [d1, d2, d3])
+        self.assertEqual(a._doc, {
+            'doc_type': 'A',
+            'l': ['2011-03-11T21:31:01Z',
+                  '2011-11-03T13:12:02Z',
+                  '2010-01-12T03:02:03Z']
+        })
+
+
+    def testListPropertyPop(self):
+        """list pop method for property w/o type
+        """
+        class A(Document):
+            l = ListProperty()
+
+        a = A()
+        a.l = [42, 24, 4224]
+        v = a.l.pop()
+        self.assert_(v == 4224)
+        self.assert_(a.l == [42, 24])
+        self.assert_(a._doc == {'doc_type': 'A', 'l': [42, 24]})
+        v = a.l.pop(0)
+        self.assert_(v == 42)
+        self.assert_(a.l == [24])
+        self.assert_(a._doc == {'doc_type': 'A', 'l': [24]})
+
+
+    def testListPropertyPopWithType(self):
+        """list pop method for property w/ type
+        """
+        from datetime import datetime
+        class A(Document):
+            l = ListProperty(item_type=datetime)
+
+        a = A()
+        d1 = datetime(2011, 3, 11, 21, 31, 1)
+        d2 = datetime(2011, 11, 3, 13, 12, 2)
+        d3 = datetime(2010, 1, 12, 3, 2, 3)
+        a.l = [d1, d2, d3]
+        v = a.l.pop()
+        self.assertEqual(v, d3)
+        self.assertEqual(a.l, [d1, d2])
+
          
     def testDictProperty(self):
         from datetime import datetime
@@ -1353,6 +1750,349 @@ class PropertyTestCase(unittest.TestCase):
         b.d = {}
         self.assert_(b.d == {})
         self.assert_(b.to_json()['d'] == {})
+
+
+
+
+class SetPropertyTestCase(unittest.TestCase):
+    def testSetPropertyConstructor(self):
+        """SetProperty constructor including default & item_type args
+        """
+        class A(Document):
+            s = SetProperty()
+        class B(Document):
+            s = SetProperty(default=set((42, 24)))
+
+        a = A()
+        self.assertEqual(a._doc, {'doc_type': 'A', 's': []})
+        b = B()
+        self.assertEqual(b._doc['doc_type'], 'B')
+        self.assertItemsEqual(b._doc['s'], [42, 24])
+        with self.assertRaises(ValueError) as cm:
+            class C(Document):
+                s = SetProperty(item_type=tuple)
+        self.assertIn(
+            "item_type <type 'tuple'> not in set([", str(cm.exception))
+
+
+    def testSetPropertyAssignment(self):
+        """SetProperty value assignment, len, in & not in
+        """
+        class A(Document):
+            s = SetProperty()
+
+        a = A()
+        a.s = set(('foo', 'bar'))
+        self.assertEqual(a.s, set(('foo', 'bar')))
+        self.assertItemsEqual(a._doc['s'], ['foo', 'bar'])
+        self.assertEqual(len(a.s), 2)
+        self.assertEqual(len(a._doc['s']), 2)
+        self.assertIn('foo', a.s)
+        self.assertIn('foo', a._doc['s'])
+        self.assertNotIn('baz', a.s)
+        self.assertNotIn('baz', a._doc['s'])
+
+
+    def testSetPropertyAssignmentWithType(self):
+        """SetProperty value assignment, len, in & not in w/ type
+        """
+        from datetime import datetime
+        class A(Document):
+            s = SetProperty(item_type=datetime)
+
+        d1 = datetime(2011, 3, 15, 17, 8, 1)
+        a = A()
+        a.s = set((d1, ))
+        self.assertEqual(a.s, set((d1, )))
+        self.assertItemsEqual(a._doc['s'], ['2011-03-15T17:08:01Z'])
+        self.assertEqual(len(a.s), 1)
+        self.assertEqual(len(a._doc['s']), 1)
+        self.assertIn(d1, a.s)
+        self.assertIn('2011-03-15T17:08:01Z', a._doc['s'])
+        self.assertNotIn(datetime(2011, 3, 16, 10, 37, 2), a.s)
+        self.assertNotIn('2011-03-16T10:37:02Z', a._doc['s'])
+
+
+    def testSetPropertySubSuperDisjoint(self):
+        """SetProperty Python subset, superset & disjoint operators work
+        """
+        class A(Document):
+            s = SetProperty()
+
+        iter1 = ('foo', 'bar')
+        a = A()
+        a.s = set(iter1)
+        self.assertTrue(a.s.issubset(iter1))
+        self.assertTrue(a.s <= set(iter1))
+        iter2 = ('foo', 'bar', 'baz')
+        self.assertTrue(a.s < set(iter2))
+        self.assertTrue(a.s.issuperset(iter1))
+        self.assertTrue(a.s >= set(iter1))
+        iter2 = ('foo', )
+        self.assertTrue(a.s > set(iter2))
+        iter2 = ('bam', 'baz')
+        self.assertTrue(a.s.isdisjoint(iter2))
+
+
+    def testSetPropertyUnionIntersectionDifferences(self):
+        """SetProperty Python union, intersection & differences operators work
+        """
+        class A(Document):
+            s = SetProperty()
+
+        iter1 = ('foo', 'bar')
+        iter2 = ('bar', 'baz')
+        iter3 = ('bar', 'fiz')
+        a = A()
+        a.s = set(iter1)
+        # Union
+        b = a.s.union(iter2)
+        self.assertEqual(b, set(('foo', 'bar', 'baz')))
+        b = a.s.union(iter2, iter3)
+        self.assertEqual(b, set(('foo', 'bar', 'baz', 'fiz')))
+        b = a.s | set(iter2)
+        self.assertEqual(b, set(('foo', 'bar', 'baz')))
+        b = a.s | set(iter2) | set(iter3)
+        self.assertEqual(b, set(('foo', 'bar', 'baz', 'fiz')))
+        # Intersection
+        b = a.s.intersection(iter2)
+        self.assertEqual(b, set(('bar', )))
+        b = a.s.intersection(iter2, iter3)
+        self.assertEqual(b, set(('bar', )))
+        b = a.s & set(iter2)
+        self.assertEqual(b, set(('bar', )))
+        b = a.s & set(iter2) & set(iter3)
+        self.assertEqual(b, set(('bar', )))
+        # Difference
+        b = a.s.difference(iter2)
+        self.assertEqual(b, set(('foo', )))
+        b = a.s.difference(iter2, iter3)
+        self.assertEqual(b, set(('foo', )))
+        b = a.s - set(iter2)
+        self.assertEqual(b, set(('foo', )))
+        b = a.s - set(iter2) - set(iter3)
+        self.assertEqual(b, set(('foo', )))
+        # Symmetric difference
+        self.assertEqual(a.s.symmetric_difference(iter2), set(('foo', 'baz')))
+        self.assertEqual(a.s ^ set(iter2), set(('foo', 'baz')))
+
+
+    def testSetPropertyCopy(self):
+        """SetProperty Python shallow copy method works
+        """
+        class A(Document):
+            s = SetProperty()
+
+        a = A()
+        a.s = set(('foo', 'bar'))
+        b = a.s.copy()
+        self.assertIsNot(b, a.s)
+
+
+    def testSetPropertyUpdate(self):
+        """SetProperty update method keeps Python set & _doc list in sync
+        """
+        class A(Document):
+            s = SetProperty()
+
+        iter1 = ('foo', 'bar')
+        iter2 = ('bar', 'baz')
+        iter3 = ('baz', 'fiz')
+        a = A()
+        a.s = set(iter1)
+        a.s.update(iter1)
+        self.assertEqual(a.s, set(('foo', 'bar')))
+        self.assertItemsEqual(a._doc['s'], ['foo', 'bar'])
+        a.s = set(iter1)
+        a.s.update(iter2)
+        self.assertEqual(a.s, set(('foo', 'bar', 'baz')))
+        self.assertItemsEqual(a._doc['s'], ['foo', 'bar', 'baz'])
+        a.s = set(iter1)
+        a.s.update(iter2, iter3)
+        self.assertEqual(a.s, set(('foo', 'bar', 'baz', 'fiz')))
+        self.assertItemsEqual(a._doc['s'], ['foo', 'bar', 'baz', 'fiz'])
+        a.s = set(iter1)
+        a.s |= set(iter2)
+        self.assertEqual(a.s, set(('foo', 'bar', 'baz')))
+        self.assertItemsEqual(a._doc['s'], ['foo', 'bar', 'baz'])
+        a.s = set(iter1)
+        a.s |= set(iter2) | set(iter3)
+        self.assertEqual(a.s, set(('foo', 'bar', 'baz', 'fiz')))
+        self.assertItemsEqual(a._doc['s'], ['foo', 'bar', 'baz', 'fiz'])
+
+
+    def testSetPropertyIntersectionUpdate(self):
+        """SetProperty intersection_update method keeps Python & _doc in sync
+        """
+        class A(Document):
+            s = SetProperty()
+
+        iter1 = ('foo', 'baz')
+        iter2 = ('bar', 'baz')
+        iter3 = ('bar', 'fiz')
+        a = A()
+        a.s = set(iter1)
+        a.s.intersection_update(iter1)
+        self.assertEqual(a.s, set(('foo', 'baz')))
+        self.assertItemsEqual(a._doc['s'], ['foo', 'baz'])
+        a.s = set(iter1)
+        a.s.intersection_update(iter2)
+        self.assertEqual(a.s, set(('baz', )))
+        self.assertItemsEqual(a._doc['s'], ['baz'])
+        a.s = set(iter1)
+        a.s.intersection_update(iter2, iter3)
+        self.assertEqual(a.s, set())
+        self.assertItemsEqual(a._doc['s'], [])
+        a.s = set(iter1)
+        a.s &= set(iter2)
+        self.assertEqual(a.s, set(('baz', )))
+        self.assertItemsEqual(a._doc['s'], ['baz'])
+        a.s = set(iter1)
+        a.s &= set(iter2) & set(iter3)
+        self.assertEqual(a.s, set())
+        self.assertItemsEqual(a._doc['s'], [])
+
+
+    def testSetPropertyDifferenceUpdate(self):
+        """SetProperty difference_update method keeps Python & _doc in sync
+        """
+        class A(Document):
+            s = SetProperty()
+
+        iter1 = ('foo', 'baz', 'fiz')
+        iter2 = ('bar', 'baz')
+        iter3 = ('bar', 'fiz')
+        a = A()
+        a.s = set(iter1)
+        a.s.difference_update(iter1)
+        self.assertEqual(a.s, set())
+        self.assertEqual(a._doc['s'], [])
+        a.s = set(iter1)
+        a.s.difference_update(iter2)
+        self.assertEqual(a.s, set(('foo', 'fiz')))
+        self.assertItemsEqual(a._doc['s'], ['foo', 'fiz'])
+        a.s = set(iter1)
+        a.s.difference_update(iter2, iter3)
+        self.assertEqual(a.s, set(('foo', )))
+        self.assertItemsEqual(a._doc['s'], ['foo'])
+        a.s = set(iter1)
+        a.s -= set(iter1)
+        self.assertEqual(a.s, set())
+        self.assertEqual(a._doc['s'], [])
+        a.s = set(iter1)
+        a.s -= set(iter2)
+        self.assertEqual(a.s, set(('foo', 'fiz')))
+        self.assertItemsEqual(a._doc['s'], ['foo', 'fiz'])
+        a.s = set(iter1)
+        a.s -= set(iter2) | set(iter3)
+        self.assertEqual(a.s, set(('foo', )))
+        self.assertItemsEqual(a._doc['s'], ['foo'])
+
+
+    def testSetPropertySymmetricDifferenceUpdate(self):
+        """SetProperty difference_update method keeps Python & _doc in sync
+        """
+        class A(Document):
+            s = SetProperty()
+
+        iter1 = ('foo', 'baz', 'fiz')
+        iter2 = ('bar', 'baz')
+        a = A()
+        a.s = set(iter1)
+        a.s.symmetric_difference_update(iter1)
+        self.assertEqual(a.s, set())
+        self.assertEqual(a._doc['s'], [])
+        a.s = set(iter1)
+        a.s.symmetric_difference_update(iter2)
+        self.assertEqual(a.s, set(('foo', 'bar', 'fiz')))
+        self.assertItemsEqual(a._doc['s'], ['foo', 'bar', 'fiz'])
+        a.s = set(iter1)
+        a.s ^= set(iter1)
+        self.assertEqual(a.s, set())
+        self.assertEqual(a._doc['s'], [])
+        a.s = set(iter1)
+        a.s ^= set(iter2)
+        self.assertEqual(a.s, set(('foo', 'bar', 'fiz')))
+        self.assertItemsEqual(a._doc['s'], ['foo', 'bar', 'fiz'])
+
+
+    def testSetPropertyAdd(self):
+        """SetProperty add method works
+        """
+        class A(Document):
+            s = SetProperty()
+
+        a = A()
+        a.s = set(('foo', 'bar'))
+        a.s.add('bar')
+        self.assertEqual(a.s, set(('foo', 'bar')))
+        self.assertItemsEqual(a._doc['s'], ['foo', 'bar'])
+        a.s.add('baz')
+        self.assertEqual(a.s, set(('foo', 'bar', 'baz')))
+        self.assertItemsEqual(a._doc['s'], ['foo', 'bar', 'baz'])
+
+
+    def testSetPropertyRemove(self):
+        """SetProperty remove method works
+        """
+        class A(Document):
+            s = SetProperty()
+
+        a = A()
+        a.s = set(('foo', 'bar'))
+        a.s.remove('foo')
+        self.assertEqual(a.s, set(('bar', )))
+        self.assertItemsEqual(a._doc['s'], ['bar'])
+        with self.assertRaises(KeyError):
+            a.s.remove('foo')
+
+
+    def testSetPropertyDiscard(self):
+        """SetProperty discard method works
+        """
+        class A(Document):
+            s = SetProperty()
+
+        a = A()
+        a.s = set(('foo', 'bar'))
+        a.s.discard('foo')
+        self.assertEqual(a.s, set(('bar', )))
+        self.assertItemsEqual(a._doc['s'], ['bar'])
+        a.s.discard('foo')
+        self.assertEqual(a.s, set(('bar', )))
+        self.assertItemsEqual(a._doc['s'], ['bar'])
+
+
+    def testSetPropertyPop(self):
+        """SetProperty pop method works
+        """
+        class A(Document):
+            s = SetProperty()
+
+        a = A()
+        a.s = set(('foo', 'bar'))
+        b = a.s.pop()
+        self.assertNotIn(b, a.s)
+        self.assertNotIn(b, a._doc['s'])
+        b = a.s.pop()
+        self.assertNotIn(b, a.s)
+        self.assertNotIn(b, a._doc['s'])
+        with self.assertRaises(KeyError):
+            a.s.pop()
+
+
+    def testSetPropertyClear(self):
+        """SetProperty clear method works
+        """
+        class A(Document):
+            s = SetProperty()
+
+        a = A()
+        a.s = set(('foo', 'bar'))
+        a.s.clear()
+        self.assertEqual(a.s, set())
+        self.assertEqual(a._doc['s'], [])
+
         
 if __name__ == '__main__':
     unittest.main()
